module advectz

  implicit none

  ! --- interface

  private

  public :: advectzzoom


contains


  subroutine advectzzoom(region)
    !
    ! set parameters for advectz
    ! written by patrick berkvens and mike botchev, march-june 1999
    !
    use dims,        only : im, jm, lm, xref, yref, zref, tref
    use dims,        only : parent, zoom2D, touch_sp, touch_np, xcyc
    use dims,        only : splitorderzoom, n_operators, status, nsplitsteps
    use toolbox,     only : escape_tm
#ifdef MPI
    use mpi_const,   only : myid,ntracetloc,my_real,com_trac,ierr,root_t,mpi_sum
    use mpi_comm,    only : barrier_t
#endif

    implicit none

    !input/output
    integer,intent(in) :: region

    ! local
    integer            :: is,ie,js,je,n,q
    integer            :: imr,jmr,lmr,tref_,xref_,yref_,zref_
    logical            :: z_encountered
    character(len=1)   :: dir
    real               :: sum_old, sum_new,sum_old_all,sum_new_all

    ! start

#ifdef MPI
    if ( ntracetloc == 0 ) return !WP! only PE's with nonzero lmloc proceed
#endif

    ! write cells along the z-walls of each child 
    ! to rm(0,...),rm(imr+1,...) and to interface cells of the child
    !
    ! this option should be there: zooming in z disabled....
    if ( .not. zoom2D ) then 
       call escape_tm('advectzzoom: ERROR zoom2D should be true:'// &
            ' zooming along z not allowed')
       !CMKCALL put_zedges(region,rm,rxm,rym,rzm,m,  &
       !CMK                        rmgl,rxmgl,rymgl,rzmgl,mgl)
    end if

    tref_ = tref(region)/tref(parent(region))
    xref_ = xref(region)/xref(parent(region))
    yref_ = yref(region)/yref(parent(region))
    zref_ = zref(region)/zref(parent(region))
    imr = im(region);  jmr = jm(region);  lmr = lm(region)

    ! determine the scope for advectz: 

    if ( region == 1 ) then
       xref_ = 0; yref_ = 0  ! to have is/ie and js/je properly computed
    end if
    ! find q - the place in the splitorderzoom 
    q=status(region)/((nsplitsteps/2)*tref_)
    ! corresponding to the begining of the last 
    ! triple x-y-z of the parent
    z_encountered=.false.

    do n=1,n_operators          ! now track following steps from q  

       dir=splitorderzoom(region,q*(nsplitsteps/2)*tref_+n)

       select case(dir)
       case('x') 
          if ( (.not.z_encountered) .or. (xcyc(region) == 1) ) then
             is=1                          ! x-substep is before z => 
             ie=im(region)                 ! full i-scope 
          else
             is=xref_+1                    ! z-substep is before x =>
             ie=im(region)-xref_           ! restricted i-scope
          end if
       case('y') 
          if (.not.z_encountered) then
             js=1                          ! y-substep is before z => 
             je=jm(region)                 ! full j-scope 
          else
             if(touch_sp(region) == 1 ) then
                js=1                       ! sp is southern edge--->full scope.
             else
                js=yref_+1                 ! x-substep is before y =>
             end if
             if ( touch_np(region) == 1 ) then
                je=jm(region)              ! np is northern edge--->full scope
             else
                je=jm(region)-yref_        ! restricted j-scope
             end if
          end if
       case('z') 
          z_encountered=.true.
       case ('c') 
       case ('v') 
       case ('s') 
       case default
          print *,'advectzzoom: strange value in splitorderzoom(',region,',',  &
               q*(nsplitsteps/2)*tref_+n,'): ',q
          call escape_tm('advectzzoom: Error in advectz')
       end select

    end do

    call advectz_work(region,is,ie,js,je)

  end subroutine advectzzoom



  subroutine advectz_work(region,is,ie,js,je)
    !
    ! calls dynamw
    ! written by mike botchev, march-june 1999
    ! modified by MK, dec 2002
    !
    use dims, only : adv_scheme

    implicit none

    ! in/out
    integer,intent(in) :: region
    integer,intent(in) :: is
    integer,intent(in) :: ie
    integer,intent(in) :: js
    integer,intent(in) :: je

    ! start

    if ( adv_scheme /= 'slope' ) &
         stop 'advectz: Wrong advection scheme: adv_scheme/=slope'

    call dynamw(region,is,ie,js,je)
    
  end subroutine advectz_work



  subroutine dynamw(region,is,ie,js,je)
    !-----------------------------------------------------------------------
    !
    !****   dynamw          - vertical tracer transport     v 9.1
    !
    ! programmed by         mh      mpi HH          23-feb-1995
    !
    !       purpose
    !       -------
    !       calculate amount of tracer moved in a vertical advection
    !       substep
    !
    !       interface
    !       ---------
    !       call dynamw
    !
    !       method
    !       ------
    !       slopes scheme
    !
    !       externals
    !       ---------
    !       none
    !
    !       reference
    !       ---------       
    !       Russel and Lerner, 1979
    !-----------------------------------------------------------------------
    !       fixed bug in calculating maximum courant number
    !                               mh, Thu, Feb 24, 1994 12:49:27
    !       changed order of loops for increased performance
    !                               mh, 11-jun-1994
    !     included code for limits of slopes to prevent negative tracer 
    !     masses                   mh, 20-jun-1994
    !
    !     zoom version written by mike botchev, march-june 1999
    !
    !-----------------------------------------------------------------------!

    use binas,       only : xgamma=>gamma
    use dims,        only : im, jm, lm, xref, yref, zref, tref
    use dims,        only : zbeg, zend, epsz, nloop_max, zero, one
    use dims,        only : okdebug, nregions, parent, limits, xi, nxi
    use global_data, only : wind_dat, mass_dat
    use chem_param,  only : ntracet, ra
    use zoom_tools,  only : mix_edges
    use toolbox,     only : escape_tm
#ifdef with_budgets    
    use budget_global, only : budget_flux, lflux1, lflux2, apply_budget_global
#endif
    use redgridZoom, only : nred, clustsize, jred, imredj
#ifdef MPI
    use mpi_const,   only : myid,root_t,my_real,com_trac,ierr, &
         mpi_max,ntracetloc,pe_first_tracer, myid, ntracet_ar
    use mpi_comm,    only : barrier_t
#endif

    implicit none

    ! input,output
    integer,intent(in) :: region
    integer,intent(in) :: is
    integer,intent(in) :: ie
    integer,intent(in) :: js
    integer,intent(in) :: je

    ! local
    real,dimension(:,:,:,:),pointer   :: rm,rxm,rym,rzm
    real,dimension(:,:,:),pointer     :: m,cm
    real,dimension(:,:,:),allocatable :: mnew, f, pf, fx, fy, cmold, mold
    real,dimension(:,:),allocatable   :: gamma
    integer                           :: i,j,l,le,lee,ls,lss,n,imr,jmr,lmr
    integer                           :: lrg
    integer                           :: tref_,xref_,yref_,zref_
    real                              :: max_one,max_all, summ
    real                              :: mxval, gam
    integer,parameter                 :: iter_limit=15
    integer                           :: n_advz, iter, nglob, offsetn, ixs, ixe, redfact

    ! start

    imr=im(region) ; jmr=jm(region) ; lmr=lm(region)

    allocate( f  ( 0:im(region)+1, 0:jm(region)+1, 0:lm(region))) 
    allocate(pf  ( 0:im(region)+1, 0:jm(region)+1, 0:lm(region))) 
    allocate(fx  ( 0:im(region)+1, 0:jm(region)+1, 0:lm(region))) 
    allocate(fy  ( 0:im(region)+1, 0:jm(region)+1, 0:lm(region))) 
    allocate(mnew( 0:im(region)+1, 0:jm(region)+1, lm(region))) 
    allocate(gamma( 0:im(region)+1, 0:jm(region)+1))
    !CMK next line: dec2003 reported by FD
    allocate(cmold(0:im(region)+1,0:jm(region)+1,0:lm(region)+1))
    allocate(mold(-1:im(region)+2,-1:jm(region)+2,lm(region)))

    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
    m => mass_dat(region)%m_t
    cm => wind_dat(region)%cm_t

    if ( okdebug ) print *,'dynamw: region=',region,' is,ie,js,je=',is,ie,js,je
    if ( ( region < 0 ) .or. ( region > nregions ) ) &
         call escape_tm( 'dynamw: STOP, illegal number of region !!!')

    ! prepare z-edges to allow for uniform work:
    !CMK not any more allowed :
    !  call compress_zedges(region,is,ie,js,je,m,rm,rxm,rym,rzm)

    ! compute refinement factors with respect to the parent
    xref_ = xref(region)/xref(parent(region))
    yref_ = yref(region)/yref(parent(region))
    zref_ = zref(region)/zref(parent(region))
    tref_ = tref(region)/tref(parent(region))
    imr=im(region);jmr=jm(region);lmr=lm(region)
    ! check is,ie,js,je:
    if ( ( is /= xref_+1 ) .and. ( is /= 1 ) ) &
         call escape_tm('dynamw: Wrong value for IS in dynamw')
    if ( ( ie /= imr-xref_ ) .and. ( ie /= imr ) ) &
         call escape_tm('dynamw: Wrong value for IE in dynamw')
    if ( ( js /= yref_+1 ) .and. ( js /= 1 ) ) &
         call escape_tm('dynamw: Wrong value for JS in dynamw')
    if ( ( je /= jmr-yref_ ) .and. ( je /= jmr ) ) &
         call escape_tm('dynamw: Wrong value for JE in dynamw')

    ! compute ls/le -- cells is:ie,js:je:ls:le will be update
    ls = zref_; le = lmr-zref_+1

    if(abs(zbeg(region)-zbeg(1))<epsz) ls = 1    ! bottom
    if(abs(zend(region)-zend(1))<epsz) le = lmr  ! top

    !
    !     if requested limit vertical slopes such that no non-negative 
    !     tracer masses should occur
    if (limits) then
      !cmk   rzm=max(min(rzm(is:ie,js:je,ls:le,:),
      !cmk         rm(is:ie,js:je,ls:le,1:ntracet)),  &
      !cmk         -rm(is:ie,js:je,ls:le,1:ntracet))
      !
      !>>> segmentation fault with ifort
      !
      !rzm(is:ie,js:je,ls:le,:) = max(min(rzm(is:ie,js:je,ls:le,:),  &
      !     rm(is:ie,js:je,ls:le,:)),  &
      !     -rm(is:ie,js:je,ls:le,:))
      !
      !<<< use do loops instead of arrays >>>
      !
      do l = ls, le
        do j = js, je
          do i = is, ie
            rzm(i,j,l,:) = max( min(rzm(i,j,l,:),rm(i,j,l,:)), -rm(i,j,l,:) )
          end do
        end do
      end do
      !
      !<<<
    end if
    !    
    !     calculate new air mass distribution
    !  
    lss = ls; lee = le    
    cmold=cm
    mold=m

    n_advz=1

    ! trial loop to determine # iterations needed:
    
    cfl: do

       !     calculate new air mass distribution
       cm = cmold/float(n_advz)
       m = mold

       ! do the iteration
       advz:   do iter=1,n_advz

          if (abs(zbeg(region)-zbeg(1))<epsz) then
             ! bottom: one-sided update of the mass
             mnew(is:ie,js:je,1)   = m(is:ie,js:je,1)   - cm(is:ie,js:je,1)
             lss = 2
          end if
          if (abs(zend(region)-zend(1))<epsz) then
             ! top:    one-sided update of the mass
             mnew(is:ie,js:je,lmr) = m(is:ie,js:je,lmr) + cm(is:ie,js:je,lmr-1)
             lee = lmr-1
          end if
          mnew(is:ie,js:je,lss:lee) = m(is:ie,js:je,lss:lee) + &
               cm(is:ie,js:je,lss-1:lee-1)  - & 
               cm(is:ie,js:je,lss:lee)

          ! reset gamma
          gam = 0.

          ! compute f, pf, fx and fy
          do l=lss-1,lee      ! 1,lmm1
             do j=1,jmr
                do i=1,imr
                   if(cm(i,j,l)>=zero) then
                      gamma(i,j)=cm(i,j,l)/m(i,j,l)
                   else
                      gamma(i,j)=cm(i,j,l)/m(i,j,l+1)
                   end if
                   xi(region,3)=max(xi(region,3),abs(gamma(i,j)))
                   if ( abs(gamma(i,j)) >= one ) then
                      nxi(region,3)=nxi(region,3)+1
                      gam=gamma(i,j)
                      exit advz
                   end if
                end do
             end do
          end do

          ! no CFLs upto now.....possibly in next?
          m(is:ie,js:je,ls:le)=mnew(is:ie,js:je,ls:le)

       end do advz

       ! if still CFL violation increase iteration counter and re-start cfl loop
       ! escape is called when the number of iterations becomes too large
       if(abs(gam)>=one) then
          if(okdebug)then
             print *,'dynamw: CFL violation: gamma,m,cm=',&
                  gam,m(i,j,l), m(i,j,l+1),cm(i,j,l),    &
                  '    in region,i,j,l,n: ',region,i,j,l,n
             print*,'dynamw: iterations:',n_advz
          end if
          n_advz = n_advz + 1

          if ( n_advz > iter_limit ) &
               call escape_tm('dynamw: too many CFL violations!')
          cycle cfl
       else
          exit cfl
       end if

    end do cfl!______CFL


    ! n_advz is now determined: apply this number of iterations:
    ! calculate new air mass distribution

    ! reset m and cm to original values:
    cm = cmold/float(n_advz)
    m = mold

    ! do the iteration
    nloop_max(region,3) = max(nloop_max(region,3), n_advz)
    do iter=1,n_advz

       if (abs(zbeg(region)-zbeg(1))<epsz) then
          ! bottom: one-sided update of the mass
          mnew(is:ie,js:je,1)   = m(is:ie,js:je,1)   - cm(is:ie,js:je,1)
          lss = 2
       end if
       if (abs(zend(region)-zend(1))<epsz) then
          ! top:    one-sided update of the mass
          mnew(is:ie,js:je,lmr) = m(is:ie,js:je,lmr) + cm(is:ie,js:je,lmr-1)
          lee = lmr-1
       end if
       mnew(is:ie,js:je,lss:lee) = m(is:ie,js:je,lss:lee) + &
            cm(is:ie,js:je,lss-1:lee-1)  - & 
            cm(is:ie,js:je,lss:lee)

       !    
       !     loop over tracers
       !    
#ifdef MPI
       do n=1,ntracetloc  !WP! changed scope
          offsetn=sum(ntracet_ar(0:myid-1))
          nglob = n + offsetn
#else   
       do n=1,ntracet
          nglob = n
#endif
          ! compute f, pf, fx and fy
          do l=lss-1,lee      ! 1,lmm1
             do j=1,jmr
                do i=1,imr
                   if(cm(i,j,l)>=zero) then
                      gamma(i,j)=cm(i,j,l)/m(i,j,l)
                      f(i,j,l)=gamma(i,j)*(rm(i,j,l,n)+(one-gamma(i,j))* &
                           rzm(i,j,l,n))
                      pf(i,j,l)=cm(i,j,l)*(gamma(i,j)*gamma(i,j)* &
                           rzm(i,j,l,n)-3.*f(i,j,l))
                      fx(i,j,l)=gamma(i,j)*rxm(i,j,l,n)
                      fy(i,j,l)=gamma(i,j)*rym(i,j,l,n)
                   else
                      gamma(i,j)=cm(i,j,l)/m(i,j,l+1)
                      f(i,j,l)=gamma(i,j)*(rm(i,j,l+1,n)-(one+gamma(i,j))* &
                           rzm(i,j,l+1,n))
                      pf(i,j,l)=cm(i,j,l)*(gamma(i,j)*gamma(i,j)* &
                           rzm(i,j,l+1,n)-3.*f(i,j,l))
                      fx(i,j,l)=gamma(i,j)*rxm(i,j,l+1,n)
                      fy(i,j,l)=gamma(i,j)*rym(i,j,l+1,n)
                   end if
                   xi(region,3)=max(xi(region,3),abs(gamma(i,j)))
                   if ( abs(gamma(i,j)) >= one ) then
                       call escape_tm('dynamw: still CFL violations?')
                   end if
                end do
             end do
          end do

          ! calculate new tracer mass, and tracer mass slopes
          ! update rm, rzm, rxm and rym in interior layers of the column

#ifdef with_budgets    
          ! calculate flux flowing in from below:
          if (apply_budget_global) then
             do j=js,je
                do i=is,ie
                   budget_flux(region)%flux_z1(i,j,nglob) = budget_flux(region)%flux_z1(i,j,nglob) +   &
                       f(i,j,lflux1(region)-1)*1e3/ra(nglob)  ! moles
                   budget_flux(region)%flux_z2(i,j,nglob) = budget_flux(region)%flux_z2(i,j,nglob) +   &
                       f(i,j,lflux2(region)-1)*1e3/ra(nglob)  ! moles
                enddo
             enddo
          endif
#endif
          do l=lss,lee      !2,lmm1
             rm(is:ie,js:je,l,n) = rm(is:ie,js:je,l,n) + &
                  f(is:ie,js:je,l-1)-f(is:ie,js:je,l)
             rzm(is:ie,js:je,l,n) = rzm(is:ie,js:je,l,n) + &
                  (pf(is:ie,js:je,l-1)-pf(is:ie,js:je,l) &
                  -(cm(is:ie,js:je,l-1)-cm(is:ie,js:je,l))* &
                  rzm(is:ie,js:je,l,n)               &
                  +3.*((cm(is:ie,js:je,l-1)+cm(is:ie,js:je,l))* &
                  rm(is:ie,js:je,l,n)            &
                  -(f(is:ie,js:je,l-1)+f(is:ie,js:je,l))* &
                  m(is:ie,js:je,l) ) ) / mnew(is:ie,js:je,l)
             if ( limits ) then    !CMK added
                rzm(is:ie,js:je,l,n) = max(min(rzm(is:ie,js:je,l,n),  &
                     rm(is:ie,js:je,l,n)),  &
                     -rm(is:ie,js:je,l,n))
             end if
             rxm(is:ie,js:je,l,n)=rxm(is:ie,js:je,l,n)+ &
                  (fx(is:ie,js:je,l-1)-fx(is:ie,js:je,l))
             rym(is:ie,js:je,l,n)=rym(is:ie,js:je,l,n)+ &
                  (fy(is:ie,js:je,l-1)-fy(is:ie,js:je,l))
          end do

          if ( abs(zbeg(region)-zbeg(1)) < epsz ) then ! bottom
             ! update rm, rzm, rxm and rym near the ground surface
             rm(is:ie,js:je,1,n)=(rm(is:ie,js:je,1,n)-f(is:ie,js:je,1))
             rzm(is:ie,js:je,1,n)=rzm(is:ie,js:je,1,n)+(-pf(is:ie,js:je,1) & 
                  +cm(is:ie,js:je,1)*rzm(is:ie,js:je,1,n) &
                  +3.*(cm(is:ie,js:je,1)*rm(is:ie,js:je,1,n) &
                  -f(is:ie,js:je,1)*m(is:ie,js:je,1)))/mnew(is:ie,js:je,1)
             if ( limits ) then   !CMK added
                rzm(is:ie,js:je,1,n) = max(min(rzm(is:ie,js:je,1,n),  &
                     rm(is:ie,js:je,1,n)),  &
                     -rm(is:ie,js:je,1,n))
             end if
             rxm(is:ie,js:je,1,n)=rxm(is:ie,js:je,1,n)-fx(is:ie,js:je,1)
             rym(is:ie,js:je,1,n)=rym(is:ie,js:je,1,n)-fy(is:ie,js:je,1)
          end if

          if (abs(zend(region)-zend(1))<epsz) then ! top
             ! update rm, rzm, rxm and rym at the top of the atmosphere
             rm(is:ie,js:je,lmr,n)=rm(is:ie,js:je,lmr,n)+f(is:ie,js:je,lmr-1)
             rzm(is:ie,js:je,lmr,n)=rzm(is:ie,js:je,lmr,n)+&
                  (pf(is:ie,js:je,lmr-1)    &
                  -cm(is:ie,js:je,lmr-1)*rzm(is:ie,js:je,lmr,n) &
                  +3.*(cm(is:ie,js:je,lmr-1)*rm(is:ie,js:je,lmr,n) &
                  -f(is:ie,js:je,lmr-1)*m(is:ie,js:je,lmr))) / &
                  mnew(is:ie,js:je,lmr)
             if ( limits ) then   !CMK added
                rzm(is:ie,js:je,lmr,n) = max(min(rzm(is:ie,js:je,lmr,n),  &
                     rm(is:ie,js:je,lmr,n)),  &
                     -rm(is:ie,js:je,lmr,n))
             end if
             rxm(is:ie,js:je,lmr,n) = rxm(is:ie,js:je,lmr,n) + &
                  fx(is:ie,js:je,lmr-1)
             rym(is:ie,js:je,lmr,n) = rym(is:ie,js:je,lmr,n) + &
                  fy(is:ie,js:je,lmr-1)
          end if

       end do ! loop over tracers

       ! store new air mass in m array
       m(is:ie,js:je,ls:le)=mnew(is:ie,js:je,ls:le)

    end do !iter

    ! finally :
    ! mix according to reduced grid (to allow for the adjoint implementation!)
    do l=1,lm(region)
       do lrg=1,nred(region)
          redfact=clustsize(lrg,region)
          j = jred(lrg,region)
          do i =  1,imredj(lrg,region)
             ixs = (i-1)*redfact + 1
             ixe = i*redfact 
             summ = sum(m(ixs:ixe,j,l))
             m(ixs:ixe,j,l) = summ/(ixe-ixs+1)
#ifdef MPI        
             do n=1,ntracetloc
#else
             do n=1,ntracet
#endif
                summ = sum(rm(ixs:ixe,j,l,n))
                rm(ixs:ixe,j,l,n) = summ/(ixe-ixs+1)
                summ = sum(rxm(ixs:ixe,j,l,n))
                rxm(ixs:ixe,j,l,n) = summ/(ixe-ixs+1)
                summ = sum(rym(ixs:ixe,j,l,n))
                rym(ixs:ixe,j,l,n) = summ/(ixe-ixs+1)
                summ = sum(rzm(ixs:ixe,j,l,n))
                rzm(ixs:ixe,j,l,n) = summ/(ixe-ixs+1)
             end do
          end do ! lrg
       end do ! j
    end do ! l


    cm=cmold    ! reset cm

    !CMK not anymore: call uncompress_zedges(region,is,ie,js,je,m,rm,rxm,rym,rzm)

    deallocate(f)
    deallocate(pf)
    deallocate(fx)
    deallocate(fy)
    deallocate(mnew)
    deallocate(gamma)
    deallocate(cmold)
    deallocate(mold)


#ifdef MPI
    if ( okdebug .and. myid == pe_first_tracer ) then
#else
    if ( okdebug ) then
#endif
       max_one=0.0
       max_one=maxval(rm(1:imr,1:jmr,1:lmr,1)/m(1:imr,1:jmr,1:lmr))
       print *, 'dynamw: z Maximum value mixing ratio',max_one
    end if
    
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(m)
    nullify(cm)
    
    call mix_edges(region)
    
  end subroutine dynamw


end module advectz
