!  Defines tracers used in inverse EnKF runs.

! Table of Contents:
!
! module chem_param
! end module chem_param

module chem_param
  !BOP
  ! !MODULE: chem_param
  ! !DESCRIPTION: 
  !  Defines tracers used by TM5.  For the inverse, there are ntrace
  ! copies of the 'co2m' tracer corresponding to the ensemble.  Note
  ! that we also define bg, ff, ocean, bio, and fire tracers with
  ! indices equal to zero.  Flux modules expect these variables to
  ! exist; we use 0 to make clear that this is an inverse run.
  !EOP

  use binas, only : xmair

  implicit none

  !BOC
  ! define specific data structures 

  ! definition of the chemistry: #reactions, order of species, etc.
  ! parameters needed for chemistry and rate constants 
  !

  ! nmark: number of 'marked' tracers 
  integer, parameter :: nmark = 1
  character(len=8),dimension(nmark) :: marknam = (/ 'ocean   ' /)

  ! ntrace: number of tracers for chemistry
  integer, parameter :: ntrace  = 150
  ! ntracet: number of transported tracers
  integer, parameter :: ntracet = 150
  ! ntrace_chem: non-transported tracers   
  integer, parameter :: ntrace_chem  = ntrace-ntracet
  ! maxtrace: total number of tracers
  integer, parameter :: maxtrace = ntrace + 0 

  !
  ! components numbers
  !
  integer, parameter :: ico2m=1  ! mean co2 mixing ratio

  ! These integers (note values are 0) defined for the sole purpose of
  ! allowing the flux modules to compile.  These tracer numbers are
  ! referenced in the "apply" routines in the modules.  For the
  ! inverse, these variables should not ever be used.  They are set to
  ! zero to cause an error in case they are accidentally used.

  integer, parameter :: ico2bg=0      ! background (initial condition) co2 mixing ratio
  integer, parameter :: ico2ff=0      ! co2 mixing ratio due to fossil fuel fluxes
  integer, parameter :: ico2ocean=0   ! co2 mixing ratio due to air-sea exchange
  integer, parameter :: ico2bio=0     ! co2 mixing ratio due to terrestrial biosphere fluxes (neutral biosphere; module without fires)
  integer, parameter :: ico2fires=0   ! co2 mixing ratio due to terrestrial biosphere fluxes (module with fires)
  !
  ! additional fields used in chemistry routine alone 
  ! (more meteo-like files in units different from #/cm3)
  !
  integer,parameter                   :: n_extra = 0
  integer,parameter                   :: nstd    = 1
  integer, dimension(nstd), parameter :: istd    = (/ 1 /)
  !
  ! species name
  !
  character(len=8), dimension(maxtrace),parameter :: names = 'co2m'
  ! 
  ! some stuff for the budgets....
  !
  integer, parameter           :: njnum = 1
  integer, parameter           :: nreac = 1
  integer, parameter           :: nhet = 1
  integer, parameter           :: nreacw = 1
  character(len=8), dimension(njnum)  :: jnam   = (/'dummy   '/)
  character(len=8), dimension(nreac)  :: ratnam = (/'dummy   '/)
  character(len=8), dimension(nreacw) :: rwnam  = (/'dummy   '/)
  character(len=8), dimension(nhet) :: rhetnam  = (/'dummy   '/)

  ! molar weights of components

  real,    parameter :: xmh=1.0079
  real,    parameter :: xmn=14.0067
  real,    parameter :: xmc=12.01115
  real,    parameter :: xms=32.064
  real,    parameter :: xmo=15.9994
  real,    parameter :: xmcl=35.453
  real,    parameter :: xmf=18.9984032
  real,    parameter :: xmh2o=xmh*2+xmo
  real,    parameter :: xmco2=xmc+2*xmo
  real,    parameter :: xmrn222=222.
  real,    parameter :: xmpb210=210.
  real,    parameter :: xmsf6= xms+6*xmf
  real,    parameter :: xmnmv=999. ! this is a dummy molecular mass

  real,dimension(maxtrace),parameter      :: ra=xmco2
  real,dimension(maxtrace),parameter      :: fscale = xmair/ra
  real,dimension(maxtrace),parameter      :: uscale = 1.e6

  !   fscale(ntrace): scaling factor for conversion of mixing ratios 
  !   in kg tracer per kg air to practical mixing ratio units (e.g. ppm)
  !   In this version: ratio of molecular weight  of tracer to that of air 
  !EOC
end module chem_param
